
TITLE 'Write CPMLDR.COM to System Track of FPGA_IDE CF Memory Card.'
; Author John Monahan	San Ramon, CA.  (10/17/2021)
;
; The CPM3 boot load process is a ONE step process (unlike the Floppy boots). The ROM monitor reads the 
; the total CPMLDR in one gulp (the first 12 (512 byte) sectors from Track 0 of the IDE Hard Disk/CF card).
; 
; >>>>>>>>>>> This is for a NON-BANKED version of CPM3 ONLY <<<<<<<<<<<<<<<<<<<<<<<<<<<
;
; The program will work with the Dual IDE/CF card board or the FPGA_DC+IDE or FPGA_DC+IDE+FDC Controller boards
;
; >>> NOTE Adjust CPM == TRUE/FALSE if this program is run from CPM or from Monitor
;  
;	10/18/2021	V1.0	Initial program
;
;
						;Equates for display on Propeller Console IO Board 
						;(Used In CPM Debugging mode only)
SCROLL			EQU	01H		;Set scrool direction UP.
LF			EQU	0AH
CR			EQU	0DH
BS			EQU	08H		;Back space (required for sector display)
BELL			EQU	07H
SPACE			EQU	20H
FAST			EQU	10H		;High speed scrool
TAB			EQU	09H		;TAB ACROSS (8 SPACES FOR SD Systems Video Board)
ESC			EQU	1BH
CLEAR			EQU	1CH		;<<< Propeller Video Board, Clear to EOL. (Use 80 spaces if EOL not available

FALSE			EQU	0
TRUE			EQU	NOT FALSE

;=================================== NOTE CPM OR MONITOR INTERFACE ==================================

CPM			EQU	TRUE		; <<<<<<<<<<<<<<<<<<<<<<<< TRUE if CRT output via CPM 
BANKED			EQU	FALSE		;<--- NOTE THIS ASSUMES WE WILL BE USING A NON-BANKED CPM3 

;===================================================================================================



STATUS$BUS$PORT		EQU	80H		;CMD Communications port with Z80 on FPGA_DC Board
DATA$BUS$PORT		EQU	81H		;CMD data is sent via this port
RELEASE$SD$PORT		EQU	82H		;A pulse out to this port will reset the SET_SD_GATE_PORT on the FPGA Z80
REBOOT$PORT		EQU	83H		;Outputting anything to this port reboots the Z80 on the FPGA_DC Board


;	Commands to the FPGA_DC Board:-
;	Note these commands are a sub-set of the ZFDC and ZFDC_II Board commands
;	Commands to the FPGA_DC Board:-
;	Note these commands are a sub-set of the old ZFDC and ZFDC_II Board commands

CMD$SEL$DRIVE		EQU	5H		;**** (Re)select an already initilized drive
CMD$SET$SIDE		EQU	8H		;**** This will set the active side for a floppy disk (only)
CMD$SET$TRK$SEC		EQU	35H		;Set new current TRACK+SECTOR on current drive (new)
CMD$READ$SECTOR		EQU	10H		;**** Read data from the CURRENT sector (on current track,drive).
CMD$WRITE$SECTOR	EQU	11H		;**** Write data to the CURRENT sector (on current track,drive).
CMD$SET$IDE$MODE	EQU	38H		;Set IDE drive hardware mode on FPG_DC Board (Internal or External)
						;Second Byte 0 = External IDE mode + Interrupt driven info (Default Mode). 
						;Second Byte 1 = External IDE mode + CMD driven info
						;Second Byte 2 = Local FPGA Board IDE mode + CMD driven info
CMD$FAST$COPY$AB	EQU	39H		;Special CMD to fast copy CF Card A:->B: (All within FPGA_DC Board software)

						
INFO$SEL$DRIVE		EQU	40H		;Send info, drive select (not currently used)
INFO$POST$MSG		EQU	41H		;Post a message on the USB Port Terminal

INFO$READ$SECTOR	EQU	42H		;Send info, read sector (Actual Sec Read done externally)
INFO$WRITE$SECTOR	EQU	43H		;Send info, write sector

CMD$BOARD$CONFIG	EQU	52H		;Get FPGA_DC Board hardware configuration

CMD$GET$TIME		EQU	80H		;Get Time and Data from FPGA DC Board RTC
CMD$SET$TIME		EQU	81H		;Set Time and Data on FPGA DC Board RTC

						;Bits 7&6 contain IDE & FDC circuit flags 
						;Bits 1&0 contain IDE Mode (0,1,2) 
						
CMD$FAILED		EQU	0FFH		;CMD failed to exacuite completely
CMD$COMPLETE		EQU	0F0H		;Flag returned by DC board to signal the previous CMD was completed OK
CMD$ABORT		EQU	0E0H		;CMD to abort the current process


;----------------------------------------------------------------------------------------------------------------------------

EXT$IDE$MODE		EQU	00H	;MODE 0, FPGA IDE mode set to external IDE ports etc, [Intererupt driven Hex display/info] 
LOCAL$IDE$MODE		EQU	01H	;MODE 1, FPGA Local IDE CMD mode   [CMD driven Hex display/info]
EXT$IDE$INFO$MODE	EQU	02H	;MODE 2, FPGA External IDE  + Info       [CMD driven Hex display/info]


IDE$BOARD		EQU	0H	;If the disk controller board is the Dual IDE/CF card board
FPGA$BOARD		EQU	1H	;If the controller is the FPGA_DC Board

IDEportA		EQU	030H		;Lower 8 bits of IDE interface (8255)
IDEportB		EQU	031H		;Upper 8 bits of IDE interface
IDEportC		EQU	032H		;Control lines for IDE interface
IDEportCtrl		EQU	033H		;8255 configuration port
IDEDrive		EQU	034H		;Bit 0 = 0 Drive A, =1 Drive B

IDE$CS$1		EQU	00000000B	; CS* select bit for IDE Card drive 1 
IDE$CS$2		EQU	00000001B	; CS* select bit for IDE Card drive 2 

READcfg8255		EQU	10010010b	;Set 8255 IDEportC to output, IDEportA/B input
WRITEcfg8255		EQU	10000000b	;Set all three 8255 ports to output mode

;IDE control lines for use with IDEportC.  

IDEa0line		EQU	01H		;direct from 8255 to IDE interface
IDEa1line		EQU	02H		;direct from 8255 to IDE interface
IDEa2line		EQU	04H		;direct from 8255 to IDE interface
IDEcs0line		EQU	08H		;inverter between 8255 and IDE interface
IDEcs1line		EQU	10H		;inverter between 8255 and IDE interface
IDEwrline		EQU	20H		;inverter between 8255 and IDE interface
IDErdline		EQU	40H		;inverter between 8255 and IDE interface
IDErstline		EQU	80H		;inverter between 8255 and IDE interface

;Symbolic constants for the IDE Drive registers, which makes the
;code more readable than always specifying the address bits

REGdata			EQU	IDEcs0line
REGerr			EQU	IDEcs0line + IDEa0line
REGseccnt		EQU	IDEcs0line + IDEa1line
REGsector		EQU	IDEcs0line + IDEa1line + IDEa0line
REGcylinderLSB		EQU	IDEcs0line + IDEa2line
REGcylinderMSB		EQU	IDEcs0line + IDEa2line + IDEa0line
REGshd			EQU	IDEcs0line + IDEa2line + IDEa1line		;(0EH)
REGcommand		EQU	IDEcs0line + IDEa2line + IDEa1line + IDEa0line	;(0FH)
REGstatus		EQU	IDEcs0line + IDEa2line + IDEa1line + IDEa0line
REGcontrol		EQU	IDEcs1line + IDEa2line + IDEa1line
REGastatus		EQU	IDEcs1line + IDEa2line + IDEa1line + IDEa0line

;IDE Command Constants.  These should never change.

COMMANDrecal		EQU	10H
COMMANDread		EQU	20H
COMMANDwrite		EQU	30H
COMMANDinit		EQU	91H
COMMANDid		EQU	0ECH
COMMANDspindown		EQU	0E0H
COMMANDspinup		EQU	0E1H


; IDE Status Register:
;  bit 7: Busy	1=busy, 0=not busy
;  bit 6: Ready 1=ready for command, 0=not ready yet
;  bit 5: DF	1=fault occured insIDE drive
;  bit 4: DSC	1=seek complete
;  bit 3: DRQ	1=data request ready, 0=not ready to xfer yet
;  bit 2: CORR	1=correctable error occured
;  bit 1: IDX	vendor specific
;  bit 0: ERR	1=error occured


				
SEC$SIZE		EQU	512	;Assume sector size as 512. (Not tested for other sizes)
MAXSEC			EQU	0FFH	;Default maximum "CPM" sector number (61, 3DH sec/track)
BUFFER$ORG		EQU	4000H	;<----- Will place all sector data here

CPM$BOOT$COUNT		EQU	12	;Allow up to 12 CPM sectors for CPMLDR

CPM$Loader$Code		EQU	2500H	;<--- Will place the CPMLDR.COM Loader here with ZSID in the HSYSGEN.SUB file

RDCON			EQU	1	;For CP/M I/O
WRCON			EQU	2
RESET$DISK		EQU	0DH	;Reset all CPM disks
PRINT			EQU	9
CONST			EQU	11	;CONSOLE STAT
BDOS			EQU	5


	; INCLUDE Z-80 MACRO LIBRARY:
	
	MACLIB	Z80			;For the Z80 DJNZ opcode




;--------------------------------------------------------------------------------------------
	ORG	100H			;<--- For CPM

START:	LXI	SP,STACK
        IF CPM
	LXI     D,SIGN$ON$CPM		;print a welcome message CPM IO
        ELSE
 	LXI     D,SIGN$ON$Z80		;print a welcome message Z80 monitor IO
        ENDIF	

	CALL	PSTRING
	
	IN	STATUS$BUS$PORT		;(80H) If FPGA_DC is active this will be X000000X
	CPI	0FFH
	JNZ	GOT$RESPONSE

	LXI	D,OLD$IDE$BOARD$MSG	;Dual IDE Board only
	CALL	PSTRING
	MVI	A,EXT$IDE$MODE		;MODE 0, for old dual IDE board
	STA	@CURRENT$IDE$MODE
	XRA	A
	STA	@BOARD$TYPE		;<<<< 0 for dual IDE Board >>>
	JP	BEGIN
	
GOT$RESPONSE:
	MVI	A,1
	STA	@BOARD$TYPE		;<<<< 1 for FPGA_DC Board >>>

;	OUT	REBOOT$PORT		;Just to be safer reboot the ZFDC_DC Z80 CPU
	LXI	D,FPGA$BOARD$MSG	;The FPGA_DC Board was detected.
	CALL	PSTRING
	
	MVI	A,CMD$BOARD$CONFIG	;52H, Get the current FPGA_DC Board configuration
	CALL	DATA$TO$FPGA
	
	CALL	GET$FPGA$TO$S100	;Wait for returned signal 

	STA	@CURRENT$BOARD$HARDEWARE ;10XXXXXX = IDE, 01XXXXXX = FDC, 11XXXXXX +IDE+FDC
	
	ANI	11000000B
	CPI	11000000B
	JZ	BOTH$CIRCUITS
	CPI	10000000B
	JZ	IDE$CIRCUIT
FDC$CIRCUIT:
	LXI	D,FDC$CIRCUIT$MSG	;FDC Board only
	CALL	PSTRING
	JR	OVER$CIRCUITS
IDE$CIRCUIT:
	LXI	D,IDE$CIRCUIT$MSG	;IDE Board only
	CALL	PSTRING
	JR	OVER$CIRCUITS
BOTH$CIRCUITS:
	LXI	D,BOTH$CIRCUITS$MSG	;IDE + FDC Board
	CALL	PSTRING

OVER$CIRCUITS:
	MVI	A,CMD$SET$IDE$MODE	;38H set IDE to External mode 
	CALL	DATA$TO$FPGA
	
	MVI	A,EXT$IDE$MODE		;0H, Select External IDE FPGA circuit for drive A: or B: access, Interrupt driven
;	MVI	A,EXT$IDE$INFO$MODE	;02H, FPGA External IDE + Info [CMD driven Hex display/info only]
;	MVI	A,LOCAL$IDE$MODE	;01H, FPGA Local IDE CMD mode   [CMD driven IDE Sec R/W (+ Hex display/info on IDE only Board))
	STA	@CURRENT$IDE$MODE	;Store in RAM

	CALL	DATA$TO$FPGA
	MVI	A,EXT$IDE$MODE		;0H, Select External IDE FPGA circuit for drive A: or B: access, Interrupt driven
;	MVI	A,EXT$IDE$INFO$MODE	;02H, FPGA External IDE + Info [CMD driven Hex display/info only]
;	MVI	A,LOCAL$IDE$MODE	;01H, FPGA Local IDE CMD mode   [CMD driven IDE Sec R/W (+ Hex display/info on IDE only Board))

					;--------------- Actual start of menu commands ------------------------------
BEGIN:	CALL	ZCRLF
	MVI	A,MAXSEC		;Initialize default max sector (assuming 0FF Sec/track)
	STA	@MAX$SEC
	MVI	A,0
	STA	@FIRST$SEC		;Initilize to first sector numbered as 0 (not 1)
	LXI	H,0
	SHLD	@TRK			;Start with Track 0,sec 0
	XRA	A
	STA	@SEC
	
	
	CALL	SEL$IDE$A		;initialize the board and first drive. If there is no drive - abort
;	JNZ	NO$A$DRIVE

	CALL	IDEinit			;<<<< (RE)Initilize IDE CF Card A:  >>>>
	JZ	DRIVE$1OK		;Ret with NZ flag set if error (probably no drive)
NO$A$DRIVE:
	LXI	D,INIT$1$ERROR	
	CALL	PSTRING
	JMP	ABORT

DRIVE$1OK:
	CALL	SEL$IDE$B		;Initialize the second drive.  If there is no drive, continue

	CALL	IDEinit			;<<<< (RE)Initilize IDE CF Card B:
	JZ	DRIVE$2OK		;Second drive initialized OK

	LXI	D,INIT$2$ERROR		;Warn about second drive initilization failure, but continue
	CALL	PSTRING
	XRA	A
	STA	@Second$DriveBPresent

	CALL	SEL$IDE$A		
;	JNZ	NO$A$DRIVE
	JMP	MAINLOOP
	
DRIVE$2OK:
	MVI	A,1
	STA	@Second$DriveBPresent
	CALL	SEL$IDE$A		
;	JNZ	NO$A$DRIVE
	JMP	MAINLOOP		;And continue


ABORT:	CALL	SEL$IDE$A		;Make sure drive A is selected.
  
  IF	CPM
	MVI	C,RESET$DISK		;Reset All disks in CPM
	CALL	BDOS		
	JMP	0H			;Jump to CP/M cold start address
  ELSE
	LXI	D,HIT$RESET		;Please Hit the Reset button
	CALL	PSTRING
	CALl	ZCRLF
HOLD:	JMP	HOLD
  ENDIF
 

ERROR:	LXI     D, msgErr		;CMD error msg
	CALL	PSTRING
	JMP	MAINLOOP
	
;-------------------------- MAIN LOOP ------------------------------------------------------------

MAINLOOP:				;Print main menu
	LXI	D,MENU$STRING$0		
	CALL	PSTRING

	LDA	@CURRENT$IDE$MODE	;Get current IDE Mode
	CPI	1
	CNZ	wrlba			;Update LBA on drive if Mode 0 or 1

	CALL	IDE$PRINT$LBA		;Display current Track,sector,head#

OVER$IDE$MENU:
	LXI	D,Prompt		;Please Entar a cmd >
	CALL	PSTRING
	
					;----------------------------------------------------
	CALL	ZCI
	CPI	ESC			;Abort if ESC
	JZ	ABORT

	CALL	upper
	
	CPI	'1'
	JZ	GET$IDE$ID		; 1, Show CF card IDs for A: & B:
	CPI	'2'			
	JZ	SELECT$CARD		; 2, Select a new card
	CPI	'3'
	JZ	FORMAT			; 3, Format current card
	CPI	'4'
	JZ	WRITE$SECTORS		; 4,  Write 12 sectors to disk, (CPM Loader)
	JMP	ERROR
	

;----------------------------- SELECT CARD/DISK SECTOR ------------------------------------------------------	

SELECT$CARD:				; "D"  Select a Card/Disk	
	CALL	SEL$CURRENT$DRIVE	; Put New drive in @CURRENT$DRIVE
	JZ	MAINLOOP
	LXI	D,BAD$SEL$DR$MSG
	CALL	PSTRING
	JMP	MAINLOOP
	

;----------------------------- WRITE 12 SECTORS FROM RAM ------------------------------------------------------	

WRITE$SECTORS: 				; Write N sectors >>>> NOTE no check is made to not overwrite 
	LXI	D,WriteN$MSG		;'Write multiple sectors from RAM to the current CF card.'
	CALL	PSTRING			

	MVI	A,CPM$BOOT$COUNT	;12 CPM Boot sectors max
	STA	@SECTOR$COUNT		;Store sector count
	
	LXI     D,msgsure		;Are you sure?
	CALL	PSTRING
	CALL	ZCI
	CALL	upper
IF CPM 
ELSE
	MOV	C,A			;Show selection
	CALL	ZCO
ENDIF	
	CPI	'Y'
	JZ	N$WR$SEC1
	JMP	MAINLOOP
	
N$WR$SEC1:	
	XRA	A			;Always Track 0,sec 1
	STA	@TRK			
	INR	A
	STA	@SEC
	
	LXI	D,ONEW$MOMENT$MSG	;'One moment while data is written'
	CALL	PSTRING

	LXI	H,CPM$Loader$Code	;Location of CPM3 Boot loader code
	SHLD	@DMA

NextWSec:
	CALL	WRITESECTOR		;Write the current CF Card Sector
	SHLD	(@DMA)			;Store for next time
	PUSH	PSW
	MVI	C,'.'			;Show progress
	CALL	ZCO
	POp	PSW
	JZ	N$WRITE$OK		;Z means the sector read was OK
	
	CALL	SHOW$SECTOR$WRITE$ERROR
	CALL	ZCRLF
	JMP	MAINLOOP

N$WRITE$OK:
	LDA	@SECTOR$COUNT
	DCR	A
	STA	@SECTOR$COUNT
	JZ	N$SEC$WRITE$DONE

	CALL	NEXT$SECT		;Point to the next sector
	JR	NextWSec
	

N$SEC$WRITE$DONE
	CALL	ZCRLF
	MVI	A,CPM$BOOT$COUNT	;Get from store sector count
	CALL	HEXOUT
	LXI	D,N$SEC$WRITE$MSG	;'Sectors read to CF Card'
	CALL	PSTRING
	LXI	H,buffer		;Point back to buffer
	SHLD	@DMA
	JMP	MAINLOOP

BAD$WR$ENTRY
	LXI	D,BAD$ENTRY$MSG		;'Bad data entry format'
	CALL	PSTRING
	CALL	ZCRLF
	LXI	H,buffer		;Point back to buffer
	SHLD	@DMA
	JMP	MAINLOOP


NEXT$SECT:				;General callable routine
	PUSH	B
	LDA	@SEC
	MOV	C,A
	LDA	@MAX$SEC
	CMP	C			;At last sector on track
	JZ	EO$TRK
	JC	EO$TRK
	MOV	A,C
	INR	A			;Otherwise, on to the next sector
	STA	@SEC
	JMP	NEXT$SEC$DONE
EO$TRK:
	LDA	@FIRST$SEC
	STA	@SEC			;Back to the first ssector
	
	LDA	@TRK
	CPI	0FFH
	JZ	END$OF$CARD	
	INR	A
	STA	@TRK
	JMP	NEXT$SEC$DONE
END$OF$CARD:
	MVI	C,BELL
	CALL	ZCO
	LDA	@MAX$SEC
	STA	@SEC			;Set to FF,MAX$SEC
	MVI	A,0FFH
	STA	@TRK
NEXT$SEC$DONE:
	POP	B
	RET	


;----------------------------- FORMAT A SD Card DRIVE  ------------------------------------------------------	
	
FORMAT:				;Format (Fill sectors with E5's for CPM directory empty)
;	LDA	@BOARD$TYPE	;<<<< 1 for FPGA_DC Board >>>
;	CPI	1
;	JZ	FPGA$FORMAT
	
	LXI	D,FORMAT$MSG	;Slower Dual IDE Board
	CALL	PSTRING
	CALL	ZCI
	CALL	upper
IF CPM 
ELSE
	MOV	C,A			;Show selection
	CALL	ZCO
ENDIF	
	CPI	'Y'
	JNZ	MAINLOOP
	
	CALL	FILL$SECTOR	;Fill buffer with 0E5's (512 of them)
	CALL	ZCRLF

NEXT1$FORMAT:
	LXI	H,buffer
	SHLD	@DMA
	CALL	WRITESECTOR	;Will return error if there was one
	JZ	main19b		;Z means the sector write was OK
	CALL	ZCRLF
	JMP	MAINLOOP
main19b:
	CALL	ZEOL		;Clear line cursor is on
;	CALL	FULL$position	;Display actual current Track,sector,head#
	CALL	ZCSTS		;Any keyboard character will stop display
	CPI	01H		;CPM Says something there
	JNZ	WRNEXTSEC1
	CALL	ZCI		;Flush character
	LXI	D,CONTINUE$MSG
	CALL	PSTRING
	CALL	ZCI
	CPI	ESC
	JZ	MAINLOOP
	CALL	ZCRLF
WRNEXTSEC1:
	LHLD	@SEC
	INX	H
	SHLD	@SEC		;0 to MAXSEC CPM Sectors
	MOV	A,L
	CPI	MAXSEC
	JNZ	NEXT1$FORMAT


	LXI	H,0		;Back to CPM sector 0
	SHLD	@SEC
	LHLD	@TRK		;Bump to next track
	INX	H
	SHLD	@TRK
	MOV	A,L		;0-FFH tracks (only)
	CPI	2		;<<<< For testing
;	CPI	MAXTRK
	JNZ	NEXT1$FORMAT	

	LXI	D,FormatDone	;Tell us we are all done.
	CALL	PSTRING
	JMP	MAINLOOP

FILL$SECTOR:	
	LXI	H,buffer		;Fill buffer with 0E5's (512 of them)
	MVI	B,0
Fill0:	MVI	A,0E5H			;<-- Sector fill character (0E5's for CPM)
	MOV	M,A
	INX	H
	MOV	M,A
	INX	H
	DJNZ	Fill0
	RET


;----------------------------- SELECT DRIVES --------------------------------------------------	


DRIVE$A:
	CALL	SEL$IDE$A		;Select Drive A
	LXI     D,SETA$MSG	
	CALL	PSTRING
	CALL	ZCRLF
	jmp	MAINLOOP

DRIVE$B:
	CALL	SEL$IDE$B		;Select Drive B
	LXI     D,SETB$MSG	
	CALL	PSTRING
	CALL	ZCRLF
	jmp	MAINLOOP


;----------------------------- GET IDE CARD ID  ------------------------------------------------------	

GET$IDE$ID:
	CALL	ZCRLF	
	CALL	driveid			;Get the drive ID info for drive 0. If there is no drive, abort
	JZ	INIT$OK1	
	LXI	D,ID$ERROR
	CALL	PSTRING
	JMP	ABORT

INIT$OK1:
	CALL	ID$PRINT		;Print the drive ID info for drive 0.  If invalid, abort
	JNC	INIT$OK2
	LXI	D,DRIVE0$IDMSG
	CALL	PSTRING
INIT$OK2:
	CALL	ZCRLF
	LDA	@Second$DriveBPresent	;Was there a second drive?
	ORA	A
	JZ	MAINLOOP		;Nope
	CALL	SEL$IDE$B			;Yes - select it.
	CALL	driveid			;Get the drive ID info for drive 1.  Continue even if none.
	JZ	INIT$OK3
	LXI	D,ID$ERROR
	CALL	PSTRING
	CALL	SEL$IDE$A
	JMP	MAINLOOP
INIT$OK3:
	CALL	ID$PRINT		;Print the drive ID info for drive 1.  Continue even if invalid.
	CALL	SEL$IDE$A
	JMP	MAINLOOP

	

;--------------------------------------------------------------------------------------
;---------------- Support Routines ----------------------------------------------------
;--------------------------------------------------------------------------------------

CHECK$BOARD$TYPE:			;Check we have a FPGA DC Board. If not abort
	LDA	@BOARD$TYPE		;1 for FPGA_DC Board >>>
	ORA	A
	RNZ
	LXI	D,BOARD$ERR$MSG		;Command only valid for FPGA DC Board
	CALL	PSTRING
	POP	PSW			;Balance up stack
	JMP	MAINLOOP

SEL$CURRENT$DRIVE:	
	LXI	D,CURRENT$DR$MSG	;Select Drive(A,B)
	CALL	PSTRING

	CALL	GETCMD			;Simple UC character Input (Note, no fancy checking)
	CPI	ESC			;Abort if ESC
	JZ	ABORT
	CALL	UPPER

RESELECT$DRIVE:				;Reselect drive in [A]
	MOV	B,A
	STA	@CURRENT$DRIVE		;Store ASCII for drive #
	
	CPI	'A'
	JZ      SEL$IDE$A
	CPI	'B'
	JZ      SEL$IDE$B
	
SEL$WITH$CMD$DRIVE:
	MVI	A,CMD$SEL$DRIVE		;05H,  Select Drive CMD, drive in [B]
	CALL	DATA$TO$FPGA
	MOV	A,B			;Get selected drive
	CALL	DATA$TO$FPGA		
	CALL	GET$FPGA$TO$S100	;Wait for returned ACK signal 
	CMP	B 
	RET				;Returns with drive in A and Z if OK


SEL$IDE$A:
	LDA	@CURRENT$IDE$MODE	;Is it the old Duel IDE/CF card interface or this new interface
	CPI	LOCAL$IDE$MODE		;Is local control of IDE required
	JZ	SEL$WITH$CMD$DRIVE	;Is CMD approach 
					;Else must be External IDE Ports (Mode 0 or 2)
	MVI	A,'A'
	STA	@CURRENT$DRIVE		;Store it as A: Drive
	MVI	A,IDE$CS$1			
	STA	@CURRENT$DRIVE$PORT	;Select drive 0 (CF Card A:)
	OUT	IDEDrive		;Port 034H
	
	MOV	D,A
	MVI	E,REGsector		;Send info to drive on IDE LED circuit (on FPGA+IDE Board)
	CALL	IDEwr8D	

	LDA	@CURRENT$IDE$MODE	;Is CMD Info also required
	CPI	EXT$IDE$INFO$MODE	
	JNZ	SELECT$A$DONE		;Must be Mode 0	

	MVI	A,CMD$SEL$DRIVE		;05H  (Re)select an already initilized CF or SD Card
	CALL	DATA$TO$FPGA

	MVI	A,'A'
	MOV	B,A			;Store for below
	CALL	DATA$TO$FPGA		;Indicate Drive A is required
	CALL	GET$FPGA$TO$S100	;Get the returned confirm data
	CMP	B
	JNZ	BAD$A$SEL$DRIVE
	
SELECT$A$DONE:				;Ret Z drive letter in A	
	XRA	A
	MVI	A,'A'
	RET

BAD$A$SEL$DRIVE:
	MVI	A,'A'
	STA	@CURRENT$DRIVE		;Store it default A: Drive
	XRA	A
	DCR	A
	RET				;RET NZ if a problem

	
SEL$IDE$B:
	LDA	@CURRENT$IDE$MODE	;Is it the old Duel IDE/CF card interface or this new interface
	CPI	LOCAL$IDE$MODE		;Is local control of IDE required
	JZ	SEL$WITH$CMD$DRIVE	;Same CMD approach as SD cards
					;Must be External IDE Ports (Mode 0 or 2)
	MVI	A,'B'
	STA	@CURRENT$DRIVE		;Store it as B: Drive
	MVI	A,IDE$CS$2			
	STA	@CURRENT$DRIVE$PORT	;Select drive 1 (CF Card B:)
	OUT	IDEDrive		;Port 034H
	
	MOV	D,A
	MVI	E,REGsector		;Send info to drive on IDE LED circuit (on FPGA+IDE Board)
	CALL	IDEwr8D	

	LDA	@CURRENT$IDE$MODE	;Is CMD Info also required
	CPI	EXT$IDE$INFO$MODE	
	JNZ	SELECT$B$DONE		;Must be Mode 0		

	MVI	A,CMD$SEL$DRIVE		;05H  (Re)select an already initilized CF or SD Card
	CALL	DATA$TO$FPGA

	MVI	A,'B'
	MOV	B,A			;Store for below
	CALL	DATA$TO$FPGA		;Indicate Drive B is required
	CALL	GET$FPGA$TO$S100	;Get the returned confirm data
	CMP	B
	CALL	BAD$B$SEL$DRIVE

SELECT$B$DONE:
	XRA	A
	MVI	A,'B'
	RET
	
BAD$B$SEL$DRIVE:
	MVI	A,'B'
	STA	@CURRENT$DRIVE		;Store it default A: Drive
	XRA	A
	DCR	A
	RET				;RET NZ if a problem



SET$LBA:				;Set a new logical block address
	LDA	@TRK			;Store current data
	STA	@TRK1
	LDA	@SEC
	STA	@SEC1
	
	CALL	GETHEX			;TRACK LOW, get 2 more HEX digits
	JNC	SKIP$CR
	CPI	CR			;If CR here then no data update
	JZ	NO$UPDATE
	JMP	BAD$FORMAT$ENTRY	
SKIP$CR:	
	STA	@TRK

	CALL	ZCI			;Format is XX,XX
	CPI	','
	JNZ	BAD$FORMAT$ENTRY

IF CPM 
ELSE
	MVI	C,','
	CALL	ZCO
ENDIF
	call	GETHEX			;SECTOR, get 2 more HEX digits
	JC	NO$UPDATE

	
	MOV	C,A
	LDA	@FIRST$SEC
	ORA	A
	JZ	IS$0$FIRST		;Sec 0 is valid
	MOV	A,C
	ORA	A
	JZ	RANGE$ERROR		;Sectors numbered 1,2,3 

IS$0$FIRST:
	LDA	@MAXSEC
	CMP	C
	JC	RANGE$ERROR
	JZ	RANGE$ERROR
	MOV	A,C
	STA	@SEC
	JMP	UPDATE$OK

RANGE$ERROR:
	CALL	PAST$MAX
	LDA	@FIRST$SEC
	STA	@SEC	
	JMP	UPDATE$OK
 
BAD$FORMAT$ENTRY:
	LXI	D,BAD$LBA$MSG
	CALL	PSTRING
NO$UPDATE:	
	LDA	@TRK1			;Store current data
	STA	@TRK
	LDA	@SEC1
	STA	@SEC
	xra	A			;Ret NZ if no new data is NOT OK
	DCR	A
	RET
	
UPDATE$OK:
	CALL	ZCRLF
	xra	A			;Ret Z if no new data is OK
	RET

PAST$MAX:
	LXI	D,SEC$RANGE$MSG		;Error Sector range is
	CALL	PSTRING
	LDA	@FIRST$SEC
	CALL	HEXOUT
	MVI	C,'-'
	CALL	ZCO
	LDA	@MAXSEC
	CALL	HEXOUT
	CALL	ZCRLF
	RET	
	
;
GETHEX:
	call	ZCI			;GET A CHARACTER, convert to UC, ECHO it
	call	UPPER	
	cpI	ESC
	jz	HEXABORT$NO$ECHO
	cpi	'/'			;check 0-9, A-F
	jc	HEXABORT$NO$ECHO
	cpi	'F'+1
	jnc	HEXABORT$NO$ECHO
IF CPM 
ELSE
	PUSH	B
	PUSH	PSW
	MOV	C,A
	CALL	ZCO
	POP	PSW
	POP	B
ENDIF
	call	ASBIN			;Convert to binary
	rlc				;Shift to high nibble
	rlc
	rlc
	rlc
	MOV	b,A			;Store it
	
	call	ZCI			;GET second CHARACTER, convert to UC, ECHO it
	call	UPPER	
	cpi	ESC
	jz	HEXABORT$NO$ECHO
	CPI	CR
	JZ	ONE$DIGIT
	
	cpi	'/'			;check 0-9, A-F
	jc	HEXABORT$NO$ECHO
	cpi	'F'+1
	jnc	HEXABORT$NO$ECHO
IF CPM 
ELSE
	PUSH	B
	PUSH	PSW
	MOV	C,A
	CALL	ZCO
	POP	PSW
	POP	B
ENDIF
	call	ASBIN			;Convert to binary
	ora	B			;add in the first digit
	ora	A			;To return NC
	ret
HEXABORT$NO$ECHO:
	scf				;Set Carry flag
	ret
;
ONE$DIGIT:				;A CR wase ntered after only 1 digit
	MOV	A,B
	RAR				;Shift to high nibble back down
	RAR
	RAR
	RAR
	ANI	0FH
	ORA	A
	RET

GET$HL$HEX:
	CALL	GETCMD			;Get a 2 characters from keyboard for H
	CPI	ESC
	JZ	HEXABORT
	CPI	'/'			;check 0-9, A-F
	JC	HEXABORT
	CPI	'F'+1
	JNC	HEXABORT
	CALL	ASBIN			;Convert to binary
	RLC				;Shift to high nibble
	RLC
	RLC
	RLC
	MOV	B,A			;Store it
	CALL	GETCMD			;Get 2nd character from keyboard & ECHO
	CPI	CR
	JZ	HEXABORT
	CPI	ESC
	JZ	HEXABORT
	CPI	'/'			;check 0-9, A-F
	JC	HEXABORT
	CPI	'F'+1
	JNC	HEXABORT
	CALL	ASBIN			;Convert to binary
	ORA	B			;Add in the first digit
	MOV	H,A

	CALL	GETCMD			;Get a 2 characters from keyboard for L
	CPI	ESC
	JZ	HEXABORT
	CPI	'/'			;check 0-9, A-F
	JC	HEXABORT
	CPI	'F'+1
	JNC	HEXABORT
	CALL	ASBIN			;Convert to binary
	RLC				;Shift to high nibble
	RLC
	RLC
	RLC
	MOV	B,A			;Store it
	CALL	GETCMD			;Get 2nd character from keyboard & ECHO
	CPI	CR
	JZ	HEXABORT
	CPI	ESC
	JZ	HEXABORT
	CPI	'/'			;check 0-9, A-F
	JC	HEXABORT
	CPI	'F'+1
	JNC	HEXABORT
	CALL	ASBIN			;Convert to binary
	ORA	B			;Add in the first digit
	MOV	L,A
	ORA	A			;To return NC
	RET
HEXABORT:
	STC				;Set Carry flag 
	RET



				
driveid:
;	LDA	@CURRENT$IDE$MODE	;Get IDE Mode
;	CPI	LOCAL$IDE$MODE		;2H
;	JNZ	ALREADY$IDE		;External mode
	
;	MVI	A,CMD$SET$IDE$MODE	;38H set IDE temporly to EXTERNAL mode 
;	CALL	DATA$TO$FPGA
;	MVI	A,EXT$IDE$MODE		;0H, IDE to External IDE mode for Drive ID
;	CALL	DATA$TO$FPGA


ALREADY$IDE:
	CALL	IDEwaitnotbusy		;Do the IDEntify drive command, and return with the buffer
					;filled with info about the drive
	RC				;If Busy return NZ
	MVI	D,COMMANDid
	MVI	E,REGcommand
	CALL	IDEwr8D			;issue the command

	CALL	IDEwaitdrq		;Wait for Busy=0, DRQ=1
	JC	SHOWerrors
	CALL	CLEAR$ID$BUFFER		;Clear ID Buffer
	
	MVI	B,0			;256 words
	LXI     H,IDbuffer		;Store data here
	CALL	FAST$RD$SEC$ID		;Get 256 words of data from REGdata port to [HL]

;	LDA	@CURRENT$IDE$MODE	;Get IDE Mode
;	CPI	LOCAL$IDE$MODE		;2H
;	JNZ	driveid$Done		;If already external, just return
	
;	MVI	A,CMD$SET$IDE$MODE	;38H set IDE back to origional local mode 
;	CALL	DATA$TO$FPGA
;	MVI	A,LOCAL$IDE$MODE	;Get Store flag
;	CALL	DATA$TO$FPGA
driveid$Done
	XRA	A
	RET				;Ret Z
	


IDE$PRINT$LBA:
	LXI	D,DRIVE$STRING		;'CF Card'
	CALL	PSTRING
	LDA	@CURRENT$DRIVE		;ASCII for DRIVE
	MOV	C,A
	CALL	ZCO
	
	LXI	D,MODE$STRING		;': IDE Mode='
	CALL	PSTRING
	LDA	@CURRENT$IDE$MODE
	ORA	A
	JZ	IS$MODE$0
	CPI	1
	JZ	IS$MODE$1
	MVI	C,'2'
	CALL	ZCO
	JR	OVER$MODE
IS$MODE$0:
	MVI	C,'0'
	CALL	ZCO
	JR	OVER$MODE
IS$MODE$1:
	MVI	C,'1'
	CALL	ZCO
OVER$MODE:	
	LXI	D,SEC$RANGE		;'(Sec Range '
	CALL	PSTRING
	LDA	@FIRST$SEC
	CALL	HEXOUT
	MVI	C,'-'			;'xx-xx
	CALL	ZCO
	LDA	@MAX$SEC
	CALL	HEXOUT
	LXI	D,TRACK$STRING		;' Track '
	CALL	PSTRING
	LDA	@TRK
	CALL	HEXOUT
	LXI	D,SECTOR$STRING		;' Sector '
	CALL	PSTRING
	LDA	@SEC
	CALL	HEXOUT
	RET

POS$MSG$STRING:				;Send character string message to the FPGA_DC Board
	MOV	C,M
	CALL	ZCO
	MOV	A,M			;UP TO THE FIRST 0H
	ORA	A
	RZ
	CALL	DATA$TO$FPGA		
	INX	H
	JR	POS$MSG$STRING


printname:			;Send text up to [B]	
	INX	H		;Text is low byte high byte format
	MOV	C,M
	CALL	ZCO	
	DCX	H
	MOV	C,M
	CALL	ZCO
	INX	H
	INX	H
	DCR	B
	JNZ	printname
	ret

ZCRLF:
	PUSH	PSW
	PUSH	BC
	MVI	C,CR
	CALL	ZCO
	MVI	C,LF
	CALL	ZCO
	POP	BC
	POP	PSW
	RET

ZEOL:				;CR and clear current line
	MVI	C,CR
	CALL	ZCO
	MVI	C,CLEAR		;Note hardware dependent, (Use 80 spaces if necessary)
	CALL	ZCO
	RET

ZCSTS:
  IF  CPM
	PUSH	B
	PUSH	D
	PUSH	H
	MVI	C,CONST
	CALL	BDOS		;Returns with 1 in [A] if character at keyboard
	POP	H
	POP	D
	POP	B
	CPI	1
	RET
  ELSE	
	IN	0H		;Get Character in [A]
	ANI	02H
	RZ
	MVI	A,01H
	ORA	A
	RET
  ENDIF
	
 
ZCO:				;Write character that is in [C]
  IF  CPM
	PUSH	PSW
	PUSH	B
	PUSH	D
	PUSH	H
	MOV	E,C
	MVI	C,WRCON
	CALL	BDOS
	POP	H
	POP	D
	POP	B
	POP	PSW
	RET
  ELSE	
	PUSH	PSW	
ZCO1:	IN   	0H		;Show Character
	ANI	04H
	JZ	ZCO1
	MOV	A,C
	OUT	1H
	POP	PSW
	RET
  ENDIF

ZCI:				;Return keyboard character in [A]
  IF CPM
	PUSH	B
	PUSH	D
	PUSH	H
	MVI	C,RDCON
	CALL	BDOS
	POP	H
	POP	D
	POP	B
	RET
  ELSE
ZCI1:	IN	0H		;Get Character in [A]
	ANI	02H
	JZ	ZCI1
	IN	01H
	RET
  ENDIF


;	Print a string in [DE] up to '$'

PSTRING:
 IF CPM
	MVI	C,PRINT
	JMP	BDOS		;PRINT MESSAGE, 
 ELSE
	PUSH	B
	PUSH	D
	PUSH	H
	XCHG
PSTRX:	MOV	A,M
	CPI	'$'
	JZ	DONEP
	MOV	C,A
	CALL	ZCO
	INX	H
	JMP	PSTRX
DONEP:	POP	H
	POP	D
	POP	B
	RET
  ENDIF


SHOWerrors:
 ;  IF	NOT DEBUG
;	ORA	A		;Set NZ flag
;	STC			;Set Carry Flag
;	RET
;   ELSE
	CALL	ZCRLF
	MVI	E,REGstatus	;Get status in status register
	CALL	IDErd8D
	MOV	A,D
	ANI	1H
	JNZ	MoreError	;Go to  REGerr register for more info
				;All OK if 01000000
	PUSH	PSW		;save for return below
	ANI	80H
	JZ	NOT7
	LXI	D,DRIVE$BUSY	;Drive Busy (bit 7) stuck high.   Status = 
	CALL	PSTRING
	JMP	DONEERR
NOT7:	ANI	40H
	JNZ	NOT6
	LXI	D,DRIVE$NOT$READY  ;Drive Not Ready (bit 6) stuck low.  Status = 
	CALL	PSTRING
	JMP	DONEERR
NOT6:	ANI	20H
	JNZ	NOT5
	LXI	D,DRIVE$WR$FAULT  ;Drive write fault.    Status =
	CALL	PSTRING
	JMP	DONEERR
NOT5	LXI	D,UNKNOWN$ERROR
	CALL	PSTRING
	JMP	DONEERR

MoreError:			;Get here if bit 0 of the status register indicted a problem
	MVI	E,REGerr	;Get error code in REGerr
	CALL	IDErd8D
	MOV	A,D
	PUSH	PSW

	ANI	10H
	JZ	NOTE4
	LXI	D,SEC$NOT$FOUND
	CALL	PSTRING
	JMP	DONEERR

NOTE4:	ANI	80H
	JZ	NOTE7
	LXI	D,BAD$BLOCK
	CALL	PSTRING
	JMP	DONEERR
NOTE7:	ANI	40H
	JZ	NOTE6
	LXI	D,UNRECOVER$ERR
	CALL	PSTRING
	JMP	DONEERR
NOTE6:	ANI	4H
	JZ	NOTE2
	LXI	D,INVALID$CMD
	CALL	PSTRING
	JMP	DONEERR
NOTE2:	ANI	2H
	JZ	NOTE1
	LXI	D,TRK0$ERR
	CALL	PSTRING
	JMP	DONEERR
NOTE1:	LXI	D,UNKNOWN$ERROR1
	CALL	PSTRING
	JMP	DONEERR

DONEERR:POP	PSW
	PUSH	PSW
	CALL	ZBITS
	CALL	ZCRLF
	POP	PSW
	ORA	A		;Set NZ flag
	STC			;Set Carry flag
	RET
;   ENDIF


;------------------------------------------------------------------
; Print a 16 bit number in RAM located @ [HL] 
; (Note Special Low Byte First. Used only for Drive ID)

printparm:
	INX	H	;Index to high byte first
	MOV	a,M
	CALL	HEXOUT
	DCX	H	;Now low byte
	MOV	a,M
	CALL	HEXOUT
	RET

HHEXOUT:				;No registers altered
	CALL	HEXOUT
	PUSH	PSW
	PUSH	B
	MVI	C,'H'
	CALL	ZCO
	POP	B
	POP	PSW
	RET
	
; Print an 8 bit number, located in [A]
HEXOUT:	PUSH	PSW			;No registers altered
	PUSH	B
	PUSH	PSW
	RRC
	RRC
	RRC
	RRC
	CALL	ZCONV
	POP	PSW
	CALL	ZCONV
	POP	B
	POP	PSW
	RET

ZCONV:	ANI	0FH		;HEX to ASCII and print it
	ADI	90H
	DAA
	ACI	40H
	DAA
	MOV	C,A
	CALL	ZCO
	RET

;DISPLAY BIT PATTERN IN [A]
ZBITS:	PUSH	PSW
	PUSH	B
	PUSH	D
	MOV	E,A		
	MVI	B,8
BQ2:	DB	0CBH,23H	;Z80 Op code for SLA A,E
	MVI	A,18H
	ADC	A
	MOV	C,A
	CALL	ZCO
	DJNZ	BQ2
	POP	D
	POP	B
	POP	PSW
	RET

GETCMD:	CALL	ZCI		;GET A CHARACTER, convert to UC, ECHO it
	CPI	ESC
	RZ			;Don't echo an ESC
	CALL	UPPER
  IF NOT CPM
	PUSH	PSW		;Save it
	PUSH	B
 	MOV	C,A
	CALL	ZCO		;Echo it
	POP	B
	POP	PSW		;get it back
 ENDIF
	RET

				;Convert LC to UC
UPPER:	CPI	'a'		;must be >= lowercase a
	RC			; else go back...
	CPI	'z'+1		;must be <= lowercase z
	RNC			; else go back...
	SUI	'a'-'A'		;subtract lowercase bias
	RET

				;ASCII TO BINARY CONVERSION ROUTINE
ASBIN:	SUI	30H 
	CPI	0AH 
	RM
	SUI	07H 
	RET


HEXDUMP:			;Print a hexdump of the data in the 512 byte buffer @[HL] (Note NOT @DMA)
	PUSH	PSW		;Save everything
	PUSH	B
	PUSH	D			 
	PUSH	H
	
	CALL	ZCRLF		;CR/LF first
	MVI	D,32		;Print 32 lines total
	MVI	B,16		;16 characters across
	SHLD	@StartLineHex	;Save the buffer location for ASCII display below
	LXI	H,0
	SHLD	@BYTE$COUNT
	
SF172:	CALL	ZCRLF
	LHLD	@BYTE$COUNT
	MOV	A,H
	CALL	HEXOUT		;Print byte count in sector
	MOV	A,L
	CALL	HEXOUT		
	PUSH	D
	LXI	D,16
	DAD	D
	POP	D
	SHLD	@BYTE$COUNT	;store for next time
	CALL	BLANK
	LHLD	@StartLineHex	;BUFFER LOCATION
	SHLD	@StartLineASCII	;Store for ASCII display below
SF175:	MOV	A,M
	CALL	LBYTE		;Display [A] on CRT/LCD
	INX	H
	DJNZ	SF175
	SHLD	@StartLineHex	;Save for next line later
	CALL	ShowAscii	;Now translate to ASCII and display
	MVI	B,16		;16 characters across for next line
	DCR	D
	JNZ	SF172		;Have we done all 32 lines
;
	CALL	ZCRLF
	POP   H			;Get back origional registers
	POP   D
	POP   B
	POP   PSW
	RET
	
ShowAscii:			;Now show as ascii info
	LHLD	@StartLineASCII
	MVI	B,16		;16 ASCII characters across
XF172:	CALL	BLANK		;send a space character
	CALL	BLANK
XF175:	MOV	A,M
	ANI	7FH
	CPI	' ' 		;FILTER OUT CONTROL CHARACTERS
	JNC	XT33
XT22:	MVI	A,'.'
XT33:	CPI	07CH
	JNC	XT22
	MOV	C,A		;SET UP TO SEND
	PUSH	B
	CALL	ZCO
	POP	B
	INX	H		;Next position in buffer
	DJNZ	XF175
	RET
;
BLANK:	PUSH	B
	PUSH	H
	MVI	C,' '
	CALL	ZCO
	POP	H
	POP	B
	RET
;
LBYTE:	PUSH	PSW
	RRC
	RRC
	RRC
	RRC
	CALL	SF598
	POP	PSW
SF598:	CALL	ZCONV
	RET

CHK$SEC:			;Compare current "CPM" sector to max "CPM" sector
	PUSH	B		;Save
	MOV	C,A		;C <- Current Sector
	LDA	@MAX$SEC	;Retrieve max sector number
	MOV	B,A		;B <- max sector number
	MOV	A,C		;Get current sector back in A for compare (and return with it in A)
	CMP	B		;Current : Max
	POP	B
	RET			;Return with compare status. (Carry => Max > Current)
;
;
;================================================================================================
;
;      IDE Drive BIOS Routines written in a format that can be used directly with CPM3 
;
;================================================================================================
;
IDEinit:
	PUSH	B		;Save used registers
	PUSH	D
				;Initialze the 8255 and drive then do a hard reset on the drive, 
	MVI	A,READcfg8255	;Config 8255 chip (10010010B), read mode on return
	OUT	IDEportCtrl	;Config 8255 chip, READ mode
				
				;Hard reset the disk drive 
				;For some reason some CF cards need to the RESET line 
				;pulsed very carefully. You may need to play around   
	MVI	A,IDErstline	;with the pulse length. Symptoms are: incorrect data comming
	OUT	IDEportC	;back from a sector read (often due to the wrong sector being read)
				;I have a (negative)pulse of 2.7uSec. (10Mz Z80, two IO wait states).
	MVI	B,20H		;Which seem to work for the 5 different CF cards I have.
ResetDelay:
	DCR	B
	JNZ	ResetDelay	;Delay (reset pulse width)
	XRA	A
	OUT	IDEportC	;No IDE control lines asserted (just bit 7 of port C)
	CALL	DELAY$SHORT	;Short Delay
	CALL	IDEwaitnotbusy	;Wait for drive

	JC	WaitInitErr
	MVI	D,11100000b	;Data for IDE SDH reg (512bytes, LBA mode,single drive,head 0000)
				;For Trk,Sec,head (non LBA) use 10100000
				;Note. Cannot get LBA mode to work with an old Seagate Medalist 6531 drive.
				;have to use the non-LBA mode. (Common for old hard disks).

	MVI	E,REGshd	;00001110,(0EH) for CS0,A2,A1,  
	CALL	IDEwr8D		;Write byte to select the MASTER device

	MVI	B,02H		;<<< Adjust delay time for hard disks to get up to speed (Currently ~ 2 seconds)
				;<<< This delay need to be much longer for actual Hard Disks, OK for CF Cards.
WaitInit:
	MVI	E,REGstatus	;Get status after initilization
	CALL	IDErd8D		;Check Status (info in [D])
	MOV	A,D
	ANI	80H
	JNZ	WaitInitL	;Need a longer wait...
	POP	D		;Restore registers
	POP	B
	RET			;Return. Well check for errors when we get back
WaitInitL:
	MVI	A,2
	CALL	DELAY$LONG	;Long delay, drive has to get up to speed
	DCR	B
	JNZ	WaitInit
	XRA	A
	DCR	A
	POP	D
	POP	B
	RET			;Return NZ. Well check for errors when we get back

;;;;;;
WaitInitErr:
	XRA	A
	DCR	A		;Return NZ (error)
	POP	D		;Restore Registers
	POP	B
	RET			;Return - check for errors there.
	
DELAY$LONG:			;Long delay (Seconds)
	STA	@DELAYStore
	PUSH	B
	LXI	B,0FFFFH	;<<< May need to adjust delay time to allow cold drive to
DELAY2:	LDA	@DELAYStore	;    get up to speed.
DELAY1:	DCR	A
	JNZ	DELAY1
	DCX	B
	MOV	A,C
	ORA	B
	JNZ	DELAY2
	POP	B
	RET


DELAY$SHORT:			;DELAY ~32 MS (DOES NOT SEEM TO BE CRITICAL)
	MVI	A,40
DELAY3:	MVI	B,0
M0:	DJNZ	M0
	DCR	A
	JNZ     DELAY3 
	RET

;-------------------------------------------------------------------------------------------------------------

EXT$READSECTOR:				;EXT$IDE$MODE Mode Sector Read @DMA. (Note @DMA is updated)
	call	SET$IDE$TRK$SEC		;Tell which sector we want to read from.
					;CPM TRK/SEC info to Drive TRK/SEC/Head
					;Send before error check so info is updated
	
	CALL	IDEwaitnotbusy		;Make sure drive is ready
	JC	SHOWerrors		;Returned with NZ set if error

	MVI	D,COMMANDread
	MVI	E,REGcommand
	CALL	IDEwr8D			;Send sec read command to drive.
	CALL	IDEwaitdrq		;wait until it's got the data
	JC	SHOWerrors
		
	LHLD  	@DMA			;DMA address
	PUSH	BC
	MVI	B,0			;Read 512 bytes to [HL] (256X2 bytes)
	MVI	C,IDEportA	

Ex$MoreRD16:
	MVI	A,REGdata		;REG regsiter address
	OUT	IDEportC	

	ORI	IDErdline		;08H+40H, Pulse RD line
	OUT	IDEportC
	
	INI				;Port A->(HL++)   (For speed, this is done 512 times)
	INR	C
	INI				;PORT B->(HL++)
	DCR	C
	MVI	A,REGdata		;Deassert RD line
	OUT	IDEportC
	DJNZ	Ex$MoreRD16

	SHLD  	@DMA			;Update the @DMA address
	POP	BC
	MVI	E,REGstatus
	CALL	IDErd8D
	MOV	A,D
	ANI	1H
	CNZ	SHOWerrors		;If error display status
	RET


;-------------------------------------------------------------------------------------------------------------


FAST$READ$IDE$SECTOR:			;Fast Read no updates, Drive,track,Sec already setup.
	CALL	IDEwaitnotbusy		;make sure drive is ready
	JC	SHOWerrors		;Returned with NZ set if error
	MVI	D,COMMANDread
	MVI	E,REGcommand
	CALL	IDEwr8D			;Send sec read command to drive.
	CALL	IDEwaitdrq		;wait until it's got the data
	JC	SHOWerrors
		
	LHLD  	@DMA			;DMA address

FAST$RD$SEC$ID:				;HL already setup
	PUSH	BC
	MVI	B,0			;Read 512 bytes to [HL] (256X2 bytes)
	MVI	C,IDEportA	

MoreRD16:
	MVI	A,REGdata		;REG regsiter address
	OUT	IDEportC	

	ORI	IDErdline		;08H+40H, Pulse RD line
	OUT	IDEportC
	
	INI				;Port A, (C)->(HL++)   (For speed, this is done 512 times)
	INR	C
	INI				;PORT B, (C)->(HL++)
	DCR	C

	MVI	A,REGdata		;Deassert RD line
	OUT	IDEportC
	DJNZ	MoreRD16

	POP	BC
	MVI	E,REGstatus
	CALL	IDErd8D
	MOV	A,D
	ANI	1H
	CNZ	SHOWerrors		;If error display status
	RET


	
;--------------------------------------------------------------------------------------------------------------------
WRITESECTOR:
	LDA	@CURRENT$IDE$MODE	;Is it the old Duel IDE/CF card interface or this new interface
	CPI	LOCAL$IDE$MODE		;Is local control of IDE required
	JZ	WRITE$WITH$CMD$SECTOR	;If local mode then send an IDE/CF Card the same as a SD Drive (Pure CMD  mode)

	LDA	@CURRENT$DRIVE		;Assume current drive is already selected
	CPI	'A'
	JZ	IDE$WR	
	CPI	'B'
	JZ	IDE$WR
	LXI	D,SEL$ERR$MSG		;'Must be IDE/CF card A: or B:'
	CALL	PSTRING
	XRA	A
	DCR	A
	RET
	
IDE$WR:					;Write a sector, specified by the 3 bytes in LBA 
					;Z on success, NZ to error routine if problem
	call	SET$IDE$TRK$SEC		;Tell which sector we want to read from.
					;CPM TRK/SEC info to Drive TRK/SEC/Head
					;Send before error check so info is updated

	LDA	@BOARD$TYPE		;<<<< 1 for FPGA_DC Board >>>
	CPI	1
	JNZ	FAST$WRITE$IDE$SECTOR

	MVI	A,INFO$WRITE$SECTOR	;42H, Send info, reading sector to FPGA_DC Z80
	CALL	DATA$TO$FPGA
	LHLD  	@DMA			;DMA address
	MOV	A,H
	CALL	DATA$TO$FPGA
	MOV	A,L
	CALL	DATA$TO$FPGA
	
	LDA	@TRK			;Send current track
	CALL	DATA$TO$FPGA
	LDA	@SEC			;Send current sector
	CALL	DATA$TO$FPGA

FAST$WRITE$IDE$SECTOR:			;Fasst Write no updates
	CALL	IDEwaitnotbusy		;make sure drive is ready
	JC	SHOWerrors

	MVI	D,COMMANDwrite
	MVI	E,REGcommand
	CALL	IDEwr8D			;tell drive to write a sector
	CALL	IDEwaitdrq		;wait unit it wants the data
	JC	SHOWerrors
	LHLD    @DMA
	MVI	B,0			;256X2 bytes

	MVI	A,WRITEcfg8255
	OUT	IDEportCtrl
	
	PUSH	BC
	MVI	C,IDEportA

WRSEC1:	MOV	A,M
	INX	H
	OUT	IDEportA	;LOW byte on A first
	MOV	A,M
	INX	H
	OUT	IDEportB	;then HIGH byte on B
	
	MVI	A,REGdata
	PUSH	PSW
	OUT	IDEportC	;Send write command
	ORI	IDEwrline	;Send WR pulse
	OUT	IDEportC
	POP	PSW
	OUT	IDEportC
	DJNZ	WRSEC1



        POP	BC
	
	MVI	A,READcfg8255		;Set 8255 back to read mode
	OUT	IDEportCtrl	

	MVI	E,REGstatus
	CALL	IDErd8D
	MOV	A,D
	ANI	1H
	CNZ	SHOWerrors	;If error display status
	RET


SHOW$SECTOR$WRITE$ERROR:
	LXI	D,SR$ERR$MSG		;'CF card sector READ error. TRK,SEC = $'
	CALL	PSTRING
	LDA	@TRK			;TRK = (8 bits)
	CALL	HEXOUT
	MVI	C,','
	CALL	ZCO
	LDA	@SEC
	CALL	HEXOUT			;XX,XX
	CALL	ZCRLF
	RET

	
WRITE$WITH$CMD$SECTOR:
	MVI	A,CMD$WRITE$SECTOR	;11H,  Send CMD_WRITE_SECTOR command 
	CALL	DATA$TO$FPGA
	
	LHLD  	@DMA			;DMA address
	MOV	A,H
	CALL	DATA$TO$FPGA
	MOV	A,L
	CALL	DATA$TO$FPGA

	LDA	@TRK			;Send Track
	CALL	DATA$TO$FPGA
	
	LDA	@SEC			;Send Sector
	CALL	DATA$TO$FPGA
	
	LHLD  	@DMA			;DMA address
	MVI	B,0			;Read 512 bytes to [HL] (256X2 bytes)

;------------------------

SD$SEC$WR$LOOP:				;<<<< LOOP SD Card Sector WRITE >>>>
	MOV	A,M
	CALL	DATA$TO$FPGA		;Send data in [A] to FPGA_DC board
	INX	H
	MOV	A,M
	CALL	DATA$TO$FPGA		;Send data in [A] to FPGA_DC board
	INX	H
	DJNZ	SD$SEC$WR$LOOP		;Get all 2X255 bytes

;-------------------------
	CALL	GET$FPGA$TO$S100	;Get CMD$CPMPLETE/ERROR FLAG
	CPI	CMD$COMPLETE		;0F0H, Flag returned by DC board to signal the previous CMD was completed OK
	JNZ	SHOW$SECTOR$WRITE$ERROR	
	XRA	A
	RET				;Ret Z if all OK


	
;--------------------------------------------------------------------------------------------------------------------


SET$IDE$TRK$SEC:			;Tell which sector we want to read from.
	MVI	A,0			;Send High TRK#
	MOV	D,A
	MVI	E,REGcylinderMSB
	CALL	IDEwr8D

	LDA	@TRK			;We have are in LBA mode so send directly
	MOV	D,A			;Send Low TRK#
	MVI	E,REGcylinderLSB
	CALL	IDEwr8D

	LDA	@SEC			;Get CPM requested sector (Note +1 Sector numbering assumed)
	MOV	D,A
	MVI	E,REGsector		;Send info to drive
	CALL	IDEwr8D

	MVI	D,1			;For now, one sector at a time
	MVI	E,REGseccnt
	CALL	IDEwr8D
	RET


wrlba:
	CALL	IDEwaitnotbusy		;Make sure drive isn't busy...
	JC	SHOWErrors		;If error, display status
	LDA	@MAX$SEC		;Are we using the special "no holes" disk format or not?
	CPI	3FH	
	JZ	wrlbaNoHoles		;Special case "no holes"
					;Else any other format (3DH or 3FFH)

wrlbaHoles:				;Write the logical block address to the drive's registers
					;Note we do not need to set the upper nibble of the LBA
					;It will always be 0 for these small drives

	LDA	@SEC			;LBA mode Low sectors go directly 
	STA	@DRIVE$SEC		;For Diagnostic Display Only
	MOV	D,A
	MVI	E,REGsector		;Send info to drive
	CALL	IDEwr8D			;Note: For drive we will have 0 - MAXSEC sectors only
				
	LDA	@TRK	
	STA	@DRIVE$TRK
	MOV	D,L			;Send Low TRK#
	MVI	E,REGcylinderLSB
	CALL	IDEwr8D

	MVI	D,0			;Send High TRK# 0 always
	MVI	E,REGcylinderMSB
	CALL	IDEwr8D

	MVI	D,1			;For now, one sector at a time
	MVI	E,REGseccnt
	CALL	IDEwr8D
	RET


wrlbaNoHoles:
					;See http://s100computers.com/My%20System%20Pages/IDE%20Board/My%20IDE%20Card.htm
					;See bottom of page "CPM installation info"
					;Starting with LBA 0 and without leaving an "holes"
	LHLD	@TRK			;Get the "CPM" requested track High & Low
	MOV	A,L			;Get Low byte of track
	RRC				;Get bottom two bits in high bits of A
	RRC
	ANI	0C0H			;Just what were the bottom two bits (now at the top)
	MOV	C,A			;Save in C
	LDA	@SEC			;Sector number in A
	ANI	03FH			;Take only bottom 6 bits, just in case
	ORA	C			;Add in top 2 bits of track
	STA	@DRIVE$SEC		;For diagnostic display only
	MOV	D,A			;Send info to the drive
	MVI	E,REGsector
	CALL	IDEwr8D

	MOV	A,L			;Get low byte of track again
	RRC				;Extract out just the top 6 bits
	RRC
	ANI	03FH
	MOV	C,A			;Save in C
	MOV	A,H			;Get high byte of track.
	RRC				;Rotate twice, leaving low 2 bits
	RRC				;In upper bits of A
	ANI	0C0H			;Mask all but the two bits we want
	ORA	C			;Add in the top 6 bits of the first track byte
	STA	@DRIVE$TRK
	MOV	D,A			;Send Low TRK#
	MVI	E,REGcylinderLSB
	CALL	IDEwr8D
	
	MOV	A,H			;Get high byte of track
	RRC				;Just the top 6 bits
	RRC
	ANI	03FH
	STA	@DRIVE$TRK+1
	MOV	D,A			;Send High TRK#
	MVI	E,REGcylinderMSB
	CALL	IDEwr8D

	MVI	D,1			;For now, one sector at a time
	MVI	E,REGseccnt
	CALL	IDEwr8D
	RET

IDEwaitnotbusy:				;ie Drive READY if 01000000
	MVI	B,0FFH
	MVI	A,0FFH			;Delay, must be above 80H for 4MHz Z80. Leave longer for slower drives
	STA	@DELAYStore
MoreWait:
	MVI	E,REGstatus		;wait for RDY bit to be set
	CALL	IDErd8D
	MOV	A,D
	ANI	11000000B
	XRI	01000000B
	JZ	DoneNotbusy
	DCR	B	
	JNZ	MoreWait	
	LDA	@DELAYStore		;Check timeout delay
	DCR	A
	STA	@DELAYStore
	JNZ	MoreWait
	STC				;Set carry to indicate an error
	ret
DoneNotBusy:
	ORA	A			;Clear carry it indicate no error
	RET

					;Wait for the drive to be ready to transfer data.
					;Returns the drive's status in Acc
IDEwaitdrq:
	MVI	B,0FFH
	MVI	A,0FFH			;Delay, must be above 80H for 4MHz Z80. Leave longer for slower drives
	STA	@DELAYStore

MoreDRQ:
	MVI	E,REGstatus		;wait for DRQ bit to be set
	CALL	IDErd8D
	MOV	A,D
	ANI	10001000B
	CPI	00001000B
	JZ	DoneDRQ
	DCR	B
	JNZ	MoreDRQ
	LDA	@DELAYStore		;Check timeout delay
	DCR	A
	STA	@DELAYStore
	JNZ	MoreDRQ
	STC				;Set carry to indicate error
	RET
DoneDRQ:
	ORA	A			;Clear carry
	RET

CLEAR$ID$BUFFER:			;Clear the ID Buffer area
	LXI	H,IDBuffer
	LXI	B,512
CLEAR2:	MVI	A,' '
	MOV	M,A
	INX	H
	DCX	B
	MOV	A,C
	ORA	B
	JNZ	CLEAR2
	
	LXI	H,IDBuffer		;Put in 0's for cylinder,heads,sectors etc
	LXI	B,14
CLEAR3:	MVI	A,0
	MOV	M,A
	INX	H
	DCX	B
	MOV	A,C
	ORA	B
	JNZ	CLEAR3
	RET




;------------------------------------ Display CF Card ID --------------------------------
;
ID$PRINT:				;Print the drive model number etc.
	LXI     H,IDbuffer + 12
	MOV	A,M			;If there are zero sectors (High byte) then something wrong
	ORA	A
	JNZ	ID$PRINT2
	INX	H
	MOV	A,M			;(Low Byte)
	ORA	A
	JNZ	ID$PRINT2		;Looks like we have a valid IDE drive
	
	LXI	D,BAD$DRIVE
	CALL	PSTRING
	STC				;Indicate error
	RET
ID$PRINT2:
	LXI	D,DRIVE$INFO
	CALL	PSTRING
	LXI     D, msgmdl	
	CALL	PSTRING
	LXI     H,IDbuffer + 54
	MVI	B,20			;character count in words
	CALL	printname		;Print [HL], [B] X 2 characters
	CALL	ZCRLF
					; print the drive's serial number
	LXI     D, msgsn
	CALL	PSTRING
	LXI     H,IDbuffer + 20
	MVI	B,10			;Character count in words
	CALL	printname
	CALL	ZCRLF
					;Print the drive's firmware revision string
	LXI     D, msgrev
	CALL	PSTRING
	LXI     H,IDbuffer + 46
	MVI	B, 4
	CALL	printname		;Character count in words
	CALL	ZCRLF

					;Tell user whether or not LBA is supported
	LXI	D,msgLBAsup1		;First part of LBA support message
	CALL	PSTRING	
	LDA	IDbuffer+98+1		;Bits 15-10 reserved, 9 LBA, 8 DMA
	ANI	02H
	JNZ	ID$PRINT3		;LBA is supported
	LXI	D,msgLBAnot		;LBA is not supported
	CALL	PSTRING
ID$PRINT3:
	LXI	D,msgLBAsup2
	CALL	PSTRING

					;Print the drive's cylinder, head, and sector specs
	LXI     D, msgcy
	CALL	PSTRING
	LXI     H,IDbuffer + 2
	CALL	printparm
	LXI     D,msghd
	CALL	PSTRING
	LXI     H,IDbuffer + 6
	CALL	printparm
	LXI     D, msgsc
	CALL	PSTRING
	LXI     H,IDbuffer + 12
	CALL	printparm
	CALL	ZCRLF
					;Default position will be first block 
	LXI	H,0
	SHLD	@SEC			;Default to Track 0, Sec 0
	SHLD	@TRK
	LXI	H,buffer		;Set DMA address to buffer
	SHLD	@DMA

	CALL	IDEinit			;For some reason this need to be here after getting the drive ID.
					;Otherwise sector #'s are off by one!
					;Comment, perhaps this was because the card wasn't really in LBA mode?
	RET

;------------------------------------------------------------------
; Low Level 8 bit R/W to the drive controller.  These are the routines that talk
; directly to the drive controller registers, via the 8255 chip.  
; Note the 16 bit I/O to the drive (which is only for SEC R/W) is done directly 
; in the routines READSECTOR & WRITESECTOR for speed reasons.
;
IDErd8D:				;READ 8 bits from IDE register in [E], return info in [D]
	MOV	A,E
	OUT	IDEportC		;drive address onto control lines

	ORI	IDErdline		;RD pulse pin (40H)
	OUT	IDEportC		;assert read pin

	IN	IDEportA
	MOV	D,A			;return with data in [D]

	MOV	A,E			;<---Ken Robbins suggestion
	OUT	IDEportC		;deassert RD pin

	XRA	A
	OUT	IDEportC		;Zero all port C lines
	ret


IDEwr8D:				;WRITE Data in [D] to IDE register in [E]
	MVI	A,WRITEcfg8255		;Set 8255 to write mode
	OUT	IDEportCtrl

	MOV	A,D			;Get data (or CMD) put it in 8255 A port
	OUT	IDEportA

	MOV	A,E			;select IDE register
	OUT	IDEportC

	ORI	IDEwrline		;lower WR line
	OUT	IDEportC
	
	MOV	A,E			;<-- Ken Robbins suggestion, raise WR line
	OUT	IDEportC		;deassert RD pin

	XRA	A			;Deselect all lines including WR line
	OUT	IDEportC

	MVI	A,READcfg8255		;Config 8255 chip, read mode on return
	OUT	IDEportCtrl
	RET


CPM$MOVE$CODE:				;This code is written to reside at 0H. Will be relocated by this
	LXI	H,BUFFER		;this program to move the boot CPMLDR to 100H in RAM (Overwriting this program)
	LXI	D,100H
	LXI	B,(12*512)
	LDIR
	NOP
	NOP
	NOP
	JMP	100H
CPM$MOVE$CODE$END:






;-----------------------------------------------------------------------------------------------------------------------

DATA$TO$FPGA:				;Send CMD/data to FPGA_DC Z80
	PUSH	PSW
DATA$TO$FPGA1:
	IN	STATUS$BUS$PORT	
;	CPI	0FFH			;0FFH if we have the Dual IDE/CF Card 
;	JZ	NO$FPGA$DC
	ANI	80H
	JRNZ	DATA$TO$FPGA1		;Wait until bit 7 is 0 (ie. any previous data has been read by FPGA_DC board Z80)
	POP	PSW
	OUT	DATA$BUS$PORT		;This will cause the FPGA_DC board to set the S100$RTS bit
					;(Seems to be required for data to arrive valid on FPGA_DC Board)
DATA$TO$FPGA2:
	IN	STATUS$BUS$PORT	
	ANI	80H
	JRNZ	DATA$TO$FPGA2		;Wait until bit 7 is 0 
	RET				;Ret NZ if error
NO$FPGA$DC
	POP	PSW
	RET
	

GET$FPGA$TO$S100:			;Get data from FPGA
	IN	STATUS$BUS$PORT		;Get S100 Status Port - Bit 0, 1 = FPGA data present, 0, no data present
	ANI	01H			;01H, Bit 0
	JZ	GET$FPGA$TO$S100	;wait for DC Board to present data 
	IN	DATA$BUS$PORT		;Return Data in [A], FPGA "hardware" will lower bit 7 of S100 Bus port 80H
	RET


;--------------------------------------------------------------------------------------------------------
		
SIGN$ON$CPM:	DB	CR,LF,LF,'WD_BOOT Program (V1.0c) '
		DB	'using CPM3 Console BIOS Routines',CR,LF
		DB	'(Make sure the Disk Controller IOBYTE bit 7 is open).',CR,LF,'$'
SIGN$ON$Z80:	DB	CR,LF,LF,'WD_BOOT Program (V1.0) '
		DB	'using the Z80 Monitor for IO',CR,LF
		DB	'(Make sure the Disk Controller IOBYTE bit 7 is open).',CR,LF
		DB	'Initilizing IDE Circuit, one moment please...',CR,LF,'$'

MENU$STRING$0:	DB  CR,LF,LF
		DB 'WD@BOOT Menu',CR,LF
		DB '1.  Show CF card IDs for A: & B: ',CR,LF
		DB '2.  Select a new card (A or B)',CR,LF
		DB '3.  Format the current card.',CR,LF
IF BANKED
		DB '4.  Write a BANKED CPM Boot Loader to current CF Card/disk.',CR,LF
ELSE
		DB '4.  Write a NON-BANKED CPM Boot Loader to current CF Card/disk.',CR,LF
ENDIF
		DB '(ESC) Quit',CR,LF,'$'

		
WriteN$MSG	DB	CR,LF,'Will write the CPM3 Boot loader to the current CF Card.'
		DB	CR,LF,'Starting at Track 0, Sector 1$'
SEL$ERR$MSG	DB	CR,LF,BELL,'Must be IDE/CF card A: or B:',CR,LF,'$'

		
ID$ERROR:	DB	'Error obtaining Drive ID.',CR,LF,'$'
INIT$DR$OK:	DB	'Drive Initilized OK.',CR,LF,LF,'$'
BAD$DRIVE:	DB	CR,LF,'Drive ID Information appears invalid. '
		DB	'(Drive possibly not present).',CR,LF,'$'
DRIVE0$IDMSG:	DB	'Drive #0 - Invalid ID - Aborting',CR,LF,'$'
DRIVE$INFO:	DB	'Drive ID Paramater Information:-',CR,LF,'$'
msgmdl:		DB	'Model: $'
msgsn:		DB	'S/N:   $'
msgrev:		DB	'Rev:   $'
msgcy:		DB	'Cylinders: $'
msghd:		DB	', Heads: $'
msgsc:		DB	', Sectors: $'
msgCPMTRK:	DB	'CPM TRK = $'
msgCPMSEC:	DB	' CPM SEC = $'
msgLBA:		DB	'  (LBA = 00$'
MSGBracket	DB	')$'
msgLBAsup1:	DB	'LBA is $'
msgLBAnot:	DB	'NOT $'
msgLBAsup2	DB	'Supported',CR,LF,'$'



DRIVE$STRING	DB	'Drive $'
MODE$STRING	DB	': IDE Mode=$'
SEC$RANGE	DB	' (Sector Range $'
TRACK$STRING	DB	') Track $'
SECTOR$STRING	DB	' Sector $'
	 
Prompt:		db	CR,LF,LF,'Please enter command > $'
msgsure:	DB	CR,LF,BELL,BELL,'Warning: this will change data on the drive, '
		DB	'are you sure? (Y or N)...$'
msgrd:		DB	CR,LF,'Sector Read OK',CR,LF,'$'
msgwr:		DB	CR,LF,'Sector Write OK',CR,LF,'$'

FPGA$BOARD$MSG DB	CR,LF,'The FPGA_DC Board was detected.$'
OLD$IDE$BOARD$MSG   DB	CR,LF,'The FPGA_DC Board was NOT detected.'
		DB	CR,LF,'But the SD cards in A: and B: can be used OK.$' 

FDC$CIRCUIT$MSG	DB	CR,LF,'FDC circuit detected on FPGA_DC Board$'
IDE$CIRCUIT$MSG	DB	CR,LF,'CF card circuit detected on FPGA_DC Board$'
BOTH$CIRCUITS$MSG DB	CR,LF,'CF card and FDC circuits detected on FPGA_DC Board$'
INIT$1$ERROR:	DB	'Initilizing of First Drive failed. Aborting Program.',BELL,CR,LF,LF,'$'
INIT$2$ERROR	DB	'Initilizing of Second Drive failed. (Possibly not present).',BELL,CR,LF,'$'
msgErr		DB	CR,LF,'Sorry, that was not a valid menu option!$'
BAD$SEL$DR$MSG	DB	CR,LF,BELL,'Select Drive Error$'
ONEW$MOMENT$MSG DB	CR,LF,'One moment please while the data is written',CR,LF,'$'
N$SEC$WRITE$MSG	DB	'H CPM3 Boot Loader Sectors were written to the CF Card.$'
BAD$ENTRY$MSG	DB	CR,LF,BELL,'Bad data entry format$'
FORMAT$MSG	DB	'FORMAT DISK. Fill all sectors with E5'
		DB	CR,LF,'Warning: this will change data on the Card, '
		DB	'are you sure? (Y or N)...$'
CONTINUE$MSG	DB	CR,LF,'To Abort enter ESC. Any other key to continue. $'
FormatDone	DB	CR,LF,'Disk Format Complete.',CR,LF,'$'
SETA$MSG	DB	CR,LF,'Current Drive is now A:$'
SETB$MSG	DB	CR,LF,'Current Drive is now B:$'
BOARD$ERR$MSG	DB	CR,LF,BELL,'Sorry that Command is valid only for the FPGA DC Board$'
CURRENT$DR$MSG	DB	CR,LF,'Select Drive SD Card A or B $'
BAD$LBA$MSG	DB	CR,LF,BELL,'Data must be in the form XX,XX (Hex).$'
SEC$RANGE$MSG	DB	CR,LF,BELL,'Error. Sector range is $'
DRIVE$BUSY	DB	'Drive Busy (bit 7) stuck high.   Status = $'
DRIVE$NOT$READY	DB	'Drive Ready (bit 6) stuck low.  Status = $'
DRIVE$WR$FAULT	DB	'Drive write fault.    Status = $'
UNKNOWN$ERROR	DB	'Unknown error in status register.   Status = $'
SEC$NOT$FOUND	DB	'Sector not found. Error Register = $'
BAD$BLOCK	DB	'Bad Sector ID.    Error Register = $'
UNRECOVER$ERR	DB	'Uncorrectable data error.  Error Register = $'
INVALID$CMD	DB	'Invalid Command. Error Register = $'
TRK0$ERR	DB	'Track Zero not found. Error Register = $'
UNKNOWN$ERROR1	DB	'Unknown Error. Error Register = $'
SR$ERR$MSG	DB	CR,LF,'Card sector READ error. TRK,SEC = $'




; -------------------------- RAM usage -----------------------------------------


RAMAREA			DB	'           RAM STORE AREA -------->'		;useful for debugging
@DMA			DW	buffer
@DMA$STORE		DW	buffer
@DRIVE$SEC		DB	0H
@DRIVE$TRK		DW	0H
@SEC			DW	0H
@TRK			DW	0H

@DRIVE1			DW	0H
@SEC1			DW	0H		
@TRK1			DW	0H
@DRIVE2			DW	0H
@SEC2			DW	0H
@TRK2			DW	0H

@StartLineHex		DW	0H
@StartLineASCII		DW	0H
@BYTE$COUNT		DW	0H
@SECTOR$COUNT		DW	0H
@SECTOR$COUNT1		DW	0H
@CURRENT$DRIVE		DB	0H
@CURRENT$DRIVE$PORT	DB	0H		;0=IDE A: 1=IDE B:
@Second$DriveBPresent	DB	0H		;1 if second drive is present
@MAX$SEC		DB	3DH		;Max "CPM" sectors/track (default 3DH for "with holes" mode)
@FIRST$SEC		DB	0H
@TRACK$COUNT		DB	0H
@DELAYSTORE		DB	0H

@CURRENT$IDE$MODE	DB	0H		;0 = old IDE Board - must do things here. 
						;1 = Extenal IDE Ports  + CMD Info
						;2 = Local FPGA IDE Ports + CMD Info
@CURRENT$BOARD$HARDEWARE DB     0H		;0XXXXXX = IDE, 01XXXXXX = FDC, 11XXXXXX +IDE+FDC
@BOARD$TYPE		DB	0H		;0= Dual IDE/CF Board, 1= FPGA_DC Board

			DS	100H
STACK:			DW	0H

		ORG	2000H			;For easy debugging will place buffers here.
BUFFER:			DS	512
IDBUFFER
			DB	'<--End of MYIDE Code              '
